import subprocess

import pytest
from brewv import cli
from click.testing import CliRunner


@pytest.fixture
def cli_runner():
    runner = CliRunner(mix_stderr=False)
    with runner.isolated_filesystem():
        yield runner


def test_it_runs_using_entrypoint():
    assert subprocess.check_output("brewv --help", shell=True) is not None


def test_it_runs_using_module():
    assert subprocess.check_output("python3 -m brewv --help", shell=True) is not None


def test_it_runs_using_cli(cli_runner):
    assert cli_runner.invoke(cli, ["--help"]).exit_code == 0



import platform

BOTTLE_FILE_SUFFIX = "bottle.tar.gz"
SYSTEM = platform.system()
IS_LINUX = SYSTEM == "Linux"
IS_MAC_OS = SYSTEM == "Darwin"

MAC_VER_TO_CODENAME = {
    "10.13": "High Sierra",
    "10.14": "Mojave",
    "10.15": "Catalina",
    "10.16": "Big Sur",
    "11": "Big Sur",
}



import logging
import os

import colorlog

from .cli import cli


def main():
    log_level = os.environ.get("BREWV_LOG_LEVEL", "INFO").upper()
    handler = logging.StreamHandler()
    handler.setFormatter(
        colorlog.ColoredFormatter(fmt="%(log_color)s[%(name)s]: %(message)s")
    )
    logging.basicConfig(handlers=[handler], level=getattr(logging, log_level))
    cli()



import errno
import hashlib
import logging
import os
import platform
import requests
import subprocess

from .constants import MAC_VER_TO_CODENAME

logger = logging.getLogger("brewv")


class HashMismatch(Exception):
    pass


def symlink_force(link_path, target):
    try:
        os.symlink(link_path, target)
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove(target)
            os.symlink(link_path, target)
        else:
            raise


def download_file(url, destination_path, sha256_verification=None):
    logger.info(f"GET {url}")
    response = requests.get(url)
    response.raise_for_status()

    # This doesn't offer much security benefit since it comes from
    # the same source as the content itself
    if "X-Checksum-Sha256" in response.headers and response.headers[
        "X-Checksum-Sha256"
    ] != sha_256(response.content):
        logger.error("SHA256 mismatch occured in transit")
        raise HashMismatch

    if sha256_verification:
        if sha256_verification == sha_256(response.content):
            logger.info("SHA256 verified successfully")
        else:
            raise HashMismatch

    with open(destination_path, "wb") as f:
        f.write(response.content)


def check_output(*args, **kwargs):
    return subprocess.check_output(*args, **kwargs).decode().strip()


def mac_ver():
    """
    macOS versions increased from 10.13, 10.14, 10.15
    until Big Sur where it got wonky. Big Sur can be 10.16 or >= 11
    I'm assuming future versions will be similar: 11.x -> 12.x -> 13.x

    This will return a normalized mac_ver, that returns the string that can
    identify the OS version
    """
    mac_ver_ = platform.mac_ver()[0].split(".")
    if int(mac_ver_[0]) >= 11:
        return mac_ver_[0]
    return ".".join(platform.mac_ver()[0].split(".")[:2])


def is_supported_mac_ver():
    return mac_ver() in MAC_VER_TO_CODENAME


def sha_256(content):
    return hashlib.sha256(content).hexdigest()



import contextlib
import json
import os
import platform
import requests
import subprocess
import sys

from requests.exceptions import HTTPError

from .constants import BOTTLE_FILE_SUFFIX, IS_MAC_OS, MAC_VER_TO_CODENAME, SYSTEM
from .util import (
    check_output,
    download_file,
    is_supported_mac_ver,
    logger,
    mac_ver,
    symlink_force,
)


class BottleNotFound(Exception):
    pass


class SourceVersionNotFound(Exception):
    pass


def _get_formula_info(formula_name):
    return json.loads(check_output(["brew", "info", "--json=v1", formula_name]))[0]


def _build_bottle_os_identifier():
    if IS_MAC_OS:
        assert is_supported_mac_ver()
        codename = MAC_VER_TO_CODENAME[mac_ver()].replace(" ", "_").lower()

        if platform.machine() == "arm64":
            return f"arm64_{codename}"
        return codename

    return f"{platform.machine()}_linux"


@contextlib.contextmanager
def _install_formula_from_tap(formula_name, version, bottle_cache_file, allow_slow):
    formula_info = _get_formula_info(formula_name)
    is_homebrew_tap = formula_info["tap"] == "homebrew/core"

    if not allow_slow and is_homebrew_tap:
        logger.error(
            "Exiting since the slow option is not enabled."
            " Enable this option to do a slow git search on the homebrew repository"
        )
        sys.exit(1)

    tap_user, tap_repo = formula_info["tap"].split("/")
    brew_repo = os.path.join(
        check_output(["brew", "--repository"]),
        f"Library/Taps/{tap_user}/homebrew-{tap_repo}",
    )

    os.chdir(brew_repo)
    logger.warning("Searching for version in git directory, this may take a while...")
    sha = check_output(
        f"git grep -F '{version}' `git rev-list master -- Formula/{formula_name}.rb`"
        f" -- Formula/{formula_name}.rb | head -n1 | cut -f1 -d':'",
        shell=True,
    )
    if not sha:
        logger.error(f"Could not find a source version for {formula_name} {version}")
        sys.exit(1)

    logger.info(f"Found a version in commit: {sha}")
    subprocess.run(
        ["git", "checkout", sha],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )

    try:
        formula_info = _get_formula_info(formula_name)
        bottle_info = (
            formula_info.get("bottle", {})
            .get("stable", {})
            .get("files", {})
            .get(_build_bottle_os_identifier(), {})
        )
        if bottle_info:
            logger.info("Found bottle in tap config")
            bottle_url = bottle_info["url"]
            bottle_sha256 = bottle_info["sha256"]
            download_file(
                bottle_url, bottle_cache_file, sha256_verification=bottle_sha256
            )

            yield bottle_cache_file
        else:
            logger.info("Did not find bottle in tap config, installing from source")
            yield formula_name
    finally:
        subprocess.run(
            ["git", "checkout", "master"],
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )


def download_bottle(bottle_file_name, destination_path):
    bottle_base_urls = dict(
        Darwin="https://homebrew.bintray.com/bottles",
        Linux="https://linuxbrew.bintray.com/bottles",
    )
    url = f"{bottle_base_urls[SYSTEM]}/{bottle_file_name}"

    try:
        download_file(url, destination_path)
    except HTTPError as e:
        if e.response.status_code < 500:
            raise BottleNotFound
        raise


def fetch_listed_bottle_versions(formula_name):
    subject = "homebrew" if IS_MAC_OS else "linuxbrew"
    url = "https://api.bintray.com/search/packages"
    params = dict(
        name=formula_name,
        subject=subject,
        repo="bottles",
    )

    response = requests.get(url, params=params)
    if not response.ok or not response.json():
        return [], None
    return response.json()[0]["versions"], response.json()[0]["latest_version"]


def switch(formula_name, version=None, allow_slow=False):
    if "/" in formula_name:
        tap, formula_name = formula_name.rsplit("/", 1)
        logger.info(f"Tapping {tap}")
        subprocess.check_output(["brew", "tap", tap])

    listed_bottle_versions, latest_version = fetch_listed_bottle_versions(formula_name)
    # Only error if there are bottles that exist
    if listed_bottle_versions:
        pretty_versions = "\n- ".join(listed_bottle_versions)
        if not version:
            logger.warning(
                f"Found the following bottle versions for {formula_name}: \n- {pretty_versions}"  # noqa
            )
            sys.exit(1)
        if version not in listed_bottle_versions:
            logger.error(
                f"{version} is not an available version for {formula_name}: \n- {pretty_versions}"  # noqa
            )
            sys.exit(1)
    elif not version:
        logger.error("Version must be given since no existing bottles were found")
        sys.exit(1)

    logger.info(f"Switching {formula_name} to version {version}")
    brew_cache_dir = check_output(["brew", "--cache"])
    bottle_cache_name = f"{formula_name}--{version}"
    bottle_cache_file = os.path.join(
        brew_cache_dir,
        f"{bottle_cache_name}.{_build_bottle_os_identifier()}.{BOTTLE_FILE_SUFFIX}",
    )

    with contextlib.ExitStack() as stack:
        if os.path.isfile(bottle_cache_file):
            logger.info(f"Found bottle in cache: {bottle_cache_file}")
            formula_installation = bottle_cache_file
        else:
            logger.info("Not in cache: finding bottle to download")
            bottle_file_name = (
                f"{formula_name}-{version}"
                f".{_build_bottle_os_identifier()}.{BOTTLE_FILE_SUFFIX}"
            )
            try:
                download_bottle(bottle_file_name, bottle_cache_file)
                formula_installation = bottle_cache_file
                # We don't have a SHA to check against
                logger.warning(
                    "Bottle successfully downloaded, but cannot verify SHA256"
                )
            except BottleNotFound:
                logger.warning(f"No bottle was found for {formula_name} {version}")
                formula_installation = stack.enter_context(
                    _install_formula_from_tap(
                        formula_name, version, bottle_cache_file, allow_slow
                    )
                )

        try:
            check_output(["brew", "unlink", formula_name], stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            pass

        try:
            subprocess.run(
                ["brew", "install", formula_installation],
                check=True,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as e:
            sys.stderr.write(e.stderr.decode())
            sys.exit(e.returncode)

        brew_prefix = check_output(["brew", "--prefix"])
        formula_cellar_path = os.path.join(
            brew_prefix, "Cellar", formula_name, version, "bin", formula_name
        )
        brew_bin_path = os.path.join(brew_prefix, "bin")
        formula_bin_to_cellar_relative_path = os.path.relpath(
            formula_cellar_path, brew_bin_path
        )

        logger.debug("Linking homebrew binary")
        symlink_force(
            formula_bin_to_cellar_relative_path,
            os.path.join(brew_bin_path, formula_name),
        )

        logger.info(f"Pinning {formula_name}")
        check_output(["brew", "pin", formula_name], stderr=subprocess.DEVNULL)



import os
from os.path import exists, join
from setuptools import find_packages, setup

base_dir = os.path.dirname(__file__)
readme_path = join(base_dir, "README.md")
if exists(readme_path):
    with open(readme_path) as stream:
        long_description = stream.read()
else:
    long_description = ""

INSTALL_REQUIRES = ("click", "colorlog", "requests")
DEV_REQUIRES = ("black", "flake8", "pytest")


setup(
    name="brew-versions",
    install_requires=INSTALL_REQUIRES,
    extras_require=dict(
        dev=DEV_REQUIRES,
    ),
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Manage different versions of homebrew packages",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Josh DM",
    url="https://github.com/lime-green/brewv",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    entry_points={
        "console_scripts": [
            "brewv = brewv:main",
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.6, <4",
    license="MIT",
    keywords=["brew", "homebrew", "homebrew version", "development", "macos", "linux"],
)



import click
import os
from shutil import which

from . import core
from .constants import IS_LINUX, IS_MAC_OS, SYSTEM
from .util import is_supported_mac_ver, logger


@click.group()
def cli():
    if not which("brew"):
        raise click.ClickException("`brew` is not installed, please install it")

    if not IS_LINUX and not IS_MAC_OS:
        raise click.ClickException(f"Your system {SYSTEM} is not supported")

    if IS_MAC_OS and not is_supported_mac_ver():
        raise click.ClickException("Your macOS version is not supported")

    os.environ["HOMEBREW_NO_AUTO_UPDATE"] = "1"


@cli.command()
@click.argument("formula_name")
@click.argument("version", required=False)
@click.option(
    "--slow/--no-slow",
    "slow",
    default=False,
    type=bool,
    help="Enable this to allow searching the HUGE homebrew repository. This takes ages",
)
def switch(formula_name, slow, version=None):
    core.switch(formula_name, version, slow)
    logger.info(
        f"Successfully switched to {formula_name} {version}",
    )
